import { useContext, useEffect, useState, useCallback } from 'preact/hooks';
import { createContext, Fragment, createElement, render } from 'preact';
import { jsx, jsxs } from 'preact/jsx-runtime';
import { createPortal } from 'preact/compat';

function mitt (n) {
  return {
    all: n = n || new Map(),
    on: function (t, e) {
      var i = n.get(t);
      i && i.push(e) || n.set(t, [e]);
    },
    off: function (t, e) {
      var i = n.get(t);
      i && i.splice(i.indexOf(e) >>> 0, 1);
    },
    emit: function (t, e) {
      (n.get(t) || []).slice().map(function (n) {
        n(e);
      }), (n.get("*") || []).slice().map(function (n) {
        n(t, e);
      });
    }
  };
}

/**
 * Flatten array, one level deep.
 *
 * @param {Array<?>} arr
 *
 * @return {Array<?>}
 */

var nativeToString = Object.prototype.toString;
var nativeHasOwnProperty = Object.prototype.hasOwnProperty;

function isUndefined(obj) {
  return obj === undefined;
}

function isDefined(obj) {
  return obj !== undefined;
}

function isNil(obj) {
  return obj == null;
}

function isArray(obj) {
  return nativeToString.call(obj) === '[object Array]';
}
/**
 * Return true, if target owns a property with the given key.
 *
 * @param {Object} target
 * @param {String} key
 *
 * @return {Boolean}
 */


function has(target, key) {
  return nativeHasOwnProperty.call(target, key);
}
/**
 * Iterate over collection; returning something
 * (non-undefined) will stop iteration.
 *
 * @param  {Array|Object} collection
 * @param  {Function} iterator
 *
 * @return {Object} return result that stopped the iteration
 */


function forEach(collection, iterator) {
  var val, result;

  if (isUndefined(collection)) {
    return;
  }

  var convertKey = isArray(collection) ? toNum : identity;

  for (var key in collection) {
    if (has(collection, key)) {
      val = collection[key];
      result = iterator(val, convertKey(key));

      if (result === false) {
        return val;
      }
    }
  }
}

function identity(arg) {
  return arg;
}

function toNum(arg) {
  return Number(arg);
}
/**
 * Sets a nested property of a given object to the specified value.
 *
 * This mutates the object and returns it.
 *
 * @param {Object} target The target of the set operation.
 * @param {(string|number)[]} path The path to the nested value.
 * @param {any} value The value to set.
 */


function set(target, path, value) {
  var currentTarget = target;
  forEach(path, function (key, idx) {
    if (key === '__proto__') {
      throw new Error('illegal key: __proto__');
    }

    var nextKey = path[idx + 1];
    var nextTarget = currentTarget[key];

    if (isDefined(nextKey) && isNil(nextTarget)) {
      nextTarget = currentTarget[key] = isNaN(+nextKey) ? {} : [];
    }

    if (isUndefined(nextKey)) {
      if (isUndefined(value)) {
        delete currentTarget[key];
      } else {
        currentTarget[key] = value;
      }
    } else {
      currentTarget = nextTarget;
    }
  });
  return target;
}

class Validator {
  validateField(field, value) {
    const {
      validate
    } = field;
    let errors = [];

    if (!validate) {
      return errors;
    }

    if (validate.pattern && value && !new RegExp(validate.pattern).test(value)) {
      errors = [...errors, `Field must match pattern ${validate.pattern}.`];
    }

    if (validate.required && typeof value === 'undefined') {
      errors = [...errors, 'Field is required.'];
    }

    if ('minLength' in validate && value && value.trim().length < validate.minLength) {
      errors = [...errors, `Field must have minimum length of ${validate.minLength}`];
    }

    if ('maxLength' in validate && value && value.trim().length > validate.maxLength) {
      errors = [...errors, `Field must have maximum length of ${validate.maxLength}`];
    }

    return errors;
  }

}

function findData(data, path) {
  for (const key of path) {
    data = data[key];

    if (data === undefined) {
      return undefined;
    }
  }

  return data;
}
function findErrors(errors, path) {
  return errors[pathStringify(path)];
}
function findFieldRenderer(renderers, type) {
  return renderers.find(renderer => {
    return renderer.type === type;
  });
}
function isRequired(field) {
  return field.required;
}
function pathParse(path) {
  if (!path) {
    return [];
  }

  return path.split('.').map(key => {
    return isNaN(parseInt(key)) ? key : parseInt(key);
  });
}
function pathsEqual(a, b) {
  return a && b && a.length === b.length && a.every((value, index) => value === b[index]);
}
function pathStringify(path) {
  if (!path) {
    return '';
  }

  return path.join('.');
}
const indices = {};

const generateIndexForType = type => {
  if (type in indices) {
    indices[type]++;
  } else {
    indices[type] = 1;
  }

  return indices[type];
};

const generateIdForType = type => {
  return `${type}${generateIndexForType(type)}`;
};
function clone(data, replacer) {
  return JSON.parse(JSON.stringify(data, replacer));
}
function importSchema(schema) {
  schema = clone(schema);
  const fields = new Map();
  importField(schema, fields);
  return {
    schema,
    fields
  };
}

function importField(field, fields, parent, index) {
  const id = generateIdForType(field.type);
  field.id = id;
  fields.set(id, field);

  if (parent) {
    field.parent = parent.id;
    field.path = parent.path ? [...parent.path, 'components', index] : ['components', index];
  } else {
    field.path = [];
  }

  if (field.components) {
    importFields(field.components, fields, field);
  }
}

function importFields(components, fields, parent) {
  components.forEach((component, index) => {
    importField(component, fields, parent, index);
  });
}

function exportSchema(schema) {
  return clone(schema, (name, value) => {
    if (['id', 'parent', 'path'].includes(name)) {
      return undefined;
    }

    return value;
  });
}

/**
 * @typedef { any } Schema
 * @typedef { { [x: string]: any } } Data
 * @typedef { { [x: string]: string[] } } Errors
 * @typedef { { [x: string]: any } } Properties
 *
 * @typedef { { data: Data, schema: Schema, properties?: Properties } } FormCoreOptions
 *
 * @typedef { { data: Data, errors: Errors, schema: Schema, properties: Properties } } State
 */

/**
 * The form core.
 */

class FormCore {
  /**
   * @constructor
   * @param { FormCoreOptions } options
   */
  constructor(options) {
    const {
      schema = {},
      data = {},
      properties = {}
    } = options;
    /**
     * @private
     */

    this.emitter = mitt();
    this.validator = new Validator();
    this.fields = new Map();
    const {
      schema: importedSchema
    } = importSchema(schema);
    /**
     * @private
     * @type {Schema}
     */

    this.initialSchema = clone(importedSchema);
    /**
     * @private
     * @type {Data}
     */

    this.initialData = clone(data);
    /**
     * @type {State}
     */

    this.state = {
      data: clone(this.initialData),
      errors: {},
      schema: clone(this.initialSchema),
      properties
    };
  }

  reset() {
    this.emitter.emit('reset');
    this.setState({
      data: clone(this.initialData),
      schema: clone(this.initialSchema),
      errors: {}
    });
  }

  submit() {
    const data = this.state.data;
    const errors = this.validateAll(data);
    const payload = clone({
      data,
      errors
    });
    this.emitter.emit('submit', payload);
    return payload;
  }
  /**
   * @param  { { path: (string|number)[], value: any } } update
   */


  update(update) {
    const {
      path,
      value
    } = update;
    const field = Array.from(this.fields.values()).find(field => pathsEqual(field.path, path));
    const fieldErrors = this.validator.validateField(field, value);
    const data = set(this.getState().data, path, value);
    const errors = set(this.getState().errors, path, fieldErrors.length ? fieldErrors : undefined);
    this.setState({
      data,
      errors
    });
  }
  /**
   * @param { any } data
   * @return { {[x: string]: string[]} } errors
   */


  validateAll(data) {
    const errors = Array.from(this.fields.values()).reduce((errors, field) => {
      const {
        path
      } = field;
      const value = findData(data, path);
      const fieldErrors = this.validator.validateField(field, value);
      return set(errors, path, fieldErrors.length ? fieldErrors : undefined);
    }, {});
    this.setState({
      errors
    });
    return this.getState().errors;
  }

  getState() {
    return clone(this.state);
  }

  setState(state) {
    this.state = { ...this.state,
      ...state
    };
    this.changed(this.state);
  }

  changed(state) {
    this.emitter.emit('changed', clone(state));
  }

  setProperty(property, value) {
    const properties = set(this.getState().properties, [property], value);
    this.setState({
      properties
    });
  }

  on(event, callback) {
    this.emitter.on(event, callback);
  }

  off(event, callback) {
    this.emitter.off(event, callback);
  }

}

const FormRenderContext = createContext({
  Empty: props => {
    return null;
  },
  Children: props => {
    return props.children;
  },
  Element: props => {
    return props.children;
  }
});

const FormContext = createContext({
  fields: new Map(),
  data: {},
  errors: {},
  properties: {},

  getFieldRenderer(type) {
    return null;
  }

});

const noop = () => false;

function FormElement(props) {
  const {
    path,
    field,
    onChange
  } = props;
  const {
    data,
    errors,
    fields,
    properties,
    getFieldRenderer
  } = useContext(FormContext);
  const {
    Element
  } = useContext(FormRenderContext);
  const Renderer = getFieldRenderer(field.type);
  const {
    id
  } = field;

  if (!Renderer) {
    throw new Error(`cannot render field <${field.type}>`);
  }

  const value = findData(data, path);
  const fieldErrors = findErrors(errors, path);
  useEffect(() => {
    fields.set(id, { ...field,
      path
    });
  }, [pathStringify(path)]);
  return jsx(Element, {
    field: field,
    children: jsx(Renderer, { ...props,
      disabled: properties.readOnly || false,
      errors: fieldErrors,
      onChange: properties.readOnly ? noop : onChange,
      value: value
    })
  });
}

/**
 * This file must not be changed or exchanged.
 *
 * @see http://bpmn.io/license for more information.
 */

function Logo() {
  return jsxs("svg", {
    xmlns: "http://www.w3.org/2000/svg",
    viewBox: "0 0 14.02 5.57",
    width: "53",
    height: "21",
    style: "vertical-align:middle",
    children: [jsx("path", {
      fill: "currentColor",
      d: "M1.88.92v.14c0 .41-.13.68-.4.8.33.14.46.44.46.86v.33c0 .61-.33.95-.95.95H0V0h.95c.65 0 .93.3.93.92zM.63.57v1.06h.24c.24 0 .38-.1.38-.43V.98c0-.28-.1-.4-.32-.4zm0 1.63v1.22h.36c.2 0 .32-.1.32-.39v-.35c0-.37-.12-.48-.4-.48H.63zM4.18.99v.52c0 .64-.31.98-.94.98h-.3V4h-.62V0h.92c.63 0 .94.35.94.99zM2.94.57v1.35h.3c.2 0 .3-.09.3-.37v-.6c0-.29-.1-.38-.3-.38h-.3zm2.89 2.27L6.25 0h.88v4h-.6V1.12L6.1 3.99h-.6l-.46-2.82v2.82h-.55V0h.87zM8.14 1.1V4h-.56V0h.79L9 2.4V0h.56v4h-.64zm2.49 2.29v.6h-.6v-.6zM12.12 1c0-.63.33-1 .95-1 .61 0 .95.37.95 1v2.04c0 .64-.34 1-.95 1-.62 0-.95-.37-.95-1zm.62 2.08c0 .28.13.39.33.39s.32-.1.32-.4V.98c0-.29-.12-.4-.32-.4s-.33.11-.33.4z"
    }), jsx("path", {
      fill: "currentColor",
      d: "M0 4.53h14.02v1.04H0zM11.08 0h.63v.62h-.63zm.63 4V1h-.63v2.98z"
    })]
  });
}

function Lightbox(props) {
  const {
    open
  } = props;

  if (!open) {
    return null;
  }

  return jsxs("div", {
    class: "fjs-powered-by-lightbox",
    style: "z-index: 100; position: fixed; top: 0; left: 0;right: 0; bottom: 0",
    children: [jsx("div", {
      class: "backdrop",
      style: "width: 100%; height: 100%; background: rgba(40 40 40 / 20%)",
      onClick: props.onBackdropClick
    }), jsxs("div", {
      class: "notice",
      style: "position: absolute; left: 50%; top: 40%; transform: translate(-50%); width: 260px; padding: 10px; background: white; box-shadow: 0  1px 4px rgba(0 0 0 / 30%); font-family: Helvetica, Arial, sans-serif; font-size: 14px; display: flex; line-height: 1.3",
      children: [jsx("a", {
        href: "https://bpmn.io",
        target: "_blank",
        rel: "noopener",
        style: "margin: 15px 20px 15px 10px; align-self: center; color: #404040",
        children: jsx(Logo, {})
      }), jsxs("span", {
        children: ["Web-based tooling for BPMN, DMN, and forms powered by ", jsx("a", {
          href: "https://bpmn.io",
          target: "_blank",
          rel: "noopener",
          children: "bpmn.io"
        }), "."]
      })]
    })]
  });
}

function Link(props) {
  return jsx("div", {
    class: "fjs-powered-by fjs-form-field",
    style: "text-align: right",
    children: jsx("a", {
      href: "https://bpmn.io",
      target: "_blank",
      rel: "noopener",
      class: "fjs-powered-by-link",
      title: "Powered by bpmn.io",
      style: "color: #404040",
      onClick: props.onClick,
      children: jsx(Logo, {})
    })
  });
}

function PoweredBy(props) {
  const [open, setOpen] = useState(false);

  function toggleOpen(open) {
    return event => {
      event.preventDefault();
      setOpen(open);
    };
  }

  return jsxs(Fragment, {
    children: [createPortal(jsx(Lightbox, {
      open: open,
      onBackdropClick: toggleOpen(false)
    }), document.body), jsx(Link, {
      onClick: toggleOpen(true)
    })]
  });
}

const noop$1 = () => {};

function Form(props) {
  const {
    onSubmit = noop$1,
    onReset = noop$1,
    onChange = noop$1,
    schema
  } = props;

  const handleSubmit = event => {
    event.preventDefault();
    onSubmit();
  };

  const handleReset = event => {
    event.preventDefault();
    onReset();
  };

  return jsxs("form", {
    class: "fjs-form",
    onSubmit: handleSubmit,
    onReset: handleReset,
    children: [jsx(FormElement, {
      field: schema,
      onChange: onChange,
      path: []
    }), jsx(PoweredBy, {})]
  });
}

function formFieldClasses(errors = [], type) {
  const classes = ['fjs-form-field'];

  if (type) {
    classes.push(`fjs-${type}`);
  }

  if (errors.length) {
    classes.push('fjs-has-errors');
  }

  return classes.join(' ');
}
function prefixId(id) {
  return `fjs-form-${id}`;
}

function ButtonRenderer(props) {
  const {
    disabled,
    field
  } = props;
  const {
    action = 'submit'
  } = field;
  return jsx("div", {
    class: formFieldClasses(),
    children: jsx("button", {
      class: "fjs-button",
      type: action,
      disabled: disabled,
      children: field.label
    })
  });
}

ButtonRenderer.create = function (options = {}) {
  const type = 'button';
  const id = generateIdForType(type);
  return {
    id,
    key: id,
    label: this.label,
    type,
    ...options
  };
};

ButtonRenderer.type = 'button';
ButtonRenderer.label = 'Button';

function DefaultRenderer(props) {
  const {
    Children,
    Empty
  } = useContext(FormRenderContext);
  let {
    field,
    path
  } = props;
  const {
    id
  } = field;
  const {
    components = []
  } = field;
  return jsxs(Children, {
    class: "fjs-vertical-layout",
    field: field,
    children: [components.map(field => {
      if (field.key) {
        path = [...path.slice(0, -1), field.key];
      } else {
        path = path.slice(0, -1);
      }

      return createElement(FormElement, { ...props,
        key: id,
        path: path,
        field: field
      });
    }), components.length ? null : jsx(Empty, {})]
  });
}

DefaultRenderer.create = function (options = {}) {
  const type = 'default';
  const id = generateIdForType(type);
  return {
    components: [],
    id,
    label: this.label,
    type,
    ...options
  };
};

DefaultRenderer.type = 'default';
DefaultRenderer.label = 'Default';

function Description(props) {
  const {
    description
  } = props;

  if (!description) {
    return null;
  }

  return jsx("div", {
    class: "fjs-form-field-description",
    children: description
  });
}

function Errors(props) {
  const {
    errors
  } = props;

  if (!errors.length) {
    return null;
  }

  return jsx("div", {
    class: "fjs-form-field-error",
    children: jsx("ul", {
      children: errors.map(error => {
        return jsx("li", {
          children: error
        });
      })
    })
  });
}

function Label(props) {
  const {
    id,
    label = 'Form Field',
    required = false
  } = props;
  return jsxs("label", {
    for: id,
    class: "fjs-form-field-label",
    children: [props.children, label, required && jsx("span", {
      class: "fjs-asterix",
      children: "*"
    })]
  });
}

function TextfieldRenderer(props) {
  const {
    disabled,
    errors = [],
    field,
    path,
    value = ''
  } = props;
  const {
    description,
    id,
    label,
    validate = {}
  } = field;
  const {
    required
  } = validate;

  const onChange = ({
    target
  }) => {
    props.onChange({
      path,
      value: target.value.length ? target.value : undefined
    });
  };

  return jsxs("div", {
    class: formFieldClasses(errors),
    children: [jsx(Label, {
      id: prefixId(id),
      label: label,
      required: required
    }), jsx("input", {
      class: "fjs-input",
      disabled: disabled,
      id: prefixId(id),
      onInput: onChange,
      type: "text",
      value: value
    }), jsx(Description, {
      description: description
    }), jsx(Errors, {
      errors: errors
    })]
  });
}

TextfieldRenderer.create = function (options = {}) {
  const type = 'textfield';
  const id = generateIdForType(type);
  return {
    id,
    key: id,
    label: this.label,
    type,
    ...options
  };
};

TextfieldRenderer.type = 'textfield';
TextfieldRenderer.label = 'Text Field';

const fields = [ButtonRenderer, // CheckboxRenderer,
// ColumnsRenderer,
DefaultRenderer, // NumberRenderer,
TextfieldRenderer]; // export { createRenderer } from './CustomRenderer';

function FormRenderer(options) {
  const {
    additionalRenderers = [],
    container,
    form
  } = options;
  const renderers = [...additionalRenderers, ...fields];

  const App = () => {
    const [state, setState] = useState(form.getState());
    const formContext = {
      get fields() {
        return form.fields;
      },

      get data() {
        return state.data;
      },

      get errors() {
        return state.errors;
      },

      get properties() {
        return state.properties;
      },

      getFieldRenderer(type) {
        return findFieldRenderer(renderers, type);
      }

    };
    form.on('changed', newState => {
      setState(newState);
    });
    const onChange = useCallback(update => form.update(update), [form]);
    const {
      properties
    } = state;
    const {
      readOnly
    } = properties;
    const onSubmit = useCallback(() => {
      if (!readOnly) {
        form.submit();
      }
    }, [form, readOnly]);
    const onReset = useCallback(() => form.reset(), [form]);
    return jsx("div", {
      class: "fjs-container",
      children: jsx(FormContext.Provider, {
        value: formContext,
        children: jsx(Form, {
          onChange: onChange,
          onSubmit: onSubmit,
          onReset: onReset,
          schema: state.schema
        })
      })
    });
  };

  render(jsx(App, {}), container);
}

/**
 * @typedef { { container: Element; schema: any; data: any; properties?: any } } FormOptions
 */

/**
 * Create a form.
 *
 * @param {FormOptions} options
 *
 * @return {FormCore}
 */

function createForm(options) {
  const {
    container,
    schema,
    data,
    properties = {}
  } = options;
  const form = new FormCore({
    schema,
    data,
    properties
  });
  new FormRenderer({
    container,
    form
  });
  return form;
}

export { ButtonRenderer, DefaultRenderer, Form, FormContext, FormCore, FormRenderContext, FormRenderer, TextfieldRenderer, clone, createForm, exportSchema, fields, findData, findErrors, findFieldRenderer, generateIdForType, importSchema, isRequired, pathParse, pathStringify, pathsEqual };
//# sourceMappingURL=index.es.js.map
